from .event import EventType
from .module import BaseBus, BusFilter, BusInject


class EventBus(BaseBus, BusFilter, BusInject):
    """
    Event bus
    """

    def __init__(self, max_concurrent_tasks: int = 10):
        BaseBus.__init__(self, max_concurrent_tasks)
        BusFilter.__init__(self)
        BusInject.__init__(self)

    async def before_emit(self, event: EventType, *args, **kwargs) -> tuple[bool, dict]:
        await BusInject.resolve(self, event, args, kwargs)
        return await BusFilter.resolve(self, event, args, kwargs), kwargs

    def clear(self):
        super().clear()
        BusFilter.clear(self)
        BusInject.clear(self)



from .event import *
from .event_bus import EventBus
from .module import *

__version__ = "0.4.0"
__author__ = "Half_nothing"

__ALL__ = [
    EnumEvent,
    AbstractEvent,
    EventType,
    EventCallback,
    SyncEventCallback,
    AsyncEventCallback,
    EventCallbackFactory,
    SyncEventCallback,
    BaseBus,
    BaseModule,
    BusFilter,
    BusInject,
    MultipleError,
    EventBus
]



from .async_event_callback import AsyncEventCallback
from .event import EnumEvent, AbstractEvent, EventType
from .event_callback import EventCallback
from .event_callback_container import EventCallbackContainer
from .event_callback_factory import EventCallbackFactory
from .sync_event_callback import SyncEventCallback

__ALL__ = [
    EnumEvent,
    AbstractEvent,
    EventType,
    EventCallback,
    SyncEventCallback,
    AsyncEventCallback,
    EventCallbackFactory,
    SyncEventCallback
]



from asyncio import iscoroutinefunction

from .async_event_callback import AsyncEventCallback
from .event_callback import EventCallback, T
from .sync_event_callback import SyncEventCallback


class EventCallbackFactory:
    """
    The factory where the EventCallback instance is created
    """

    @staticmethod
    def create(callback: T, weight: int = 1) -> EventCallback:
        if iscoroutinefunction(callback):
            return AsyncEventCallback(callback, weight)
        else:
            return SyncEventCallback(callback, weight)



from abc import ABC, abstractmethod

from ..event import EventType


class BaseModule(ABC):
    """
    Event module base class
    """

    @abstractmethod
    async def resolve(self, event: EventType, args, kwargs) -> bool:
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError



from abc import ABC
from enum import Enum
from typing import Type, Union


# Enum event
class EnumEvent(Enum):
    pass


# Event class there is more room for customization
class AbstractEvent(ABC):
    pass


EventType: Type = Union[EnumEvent, Type[AbstractEvent], str]



from typing import Any, Awaitable, Callable, Type, Union

from loguru import logger

from .base_module import BaseModule
from ..event import EventType, EventCallbackContainer

InjectCallback: Type = Callable[..., Union[dict[str, Any], Awaitable[dict[str, Any]]]]


class BusInject(BaseModule):
    """
    The event bus injector module is responsible for injecting parameters into events to achieve special operations,
    such as user authentication
    """

    def __init__(self):
        self._injects: dict[str, EventCallbackContainer] = {}
        self._global_injects: EventCallbackContainer = EventCallbackContainer()

    def clear(self) -> None:
        self._injects.clear()
        self._global_injects.clear()

    async def resolve(self, event: EventType, args: tuple, kwargs: dict[str, Any]) -> bool:
        kwargs.update(await self._apply_global_injects(*args, **kwargs))
        kwargs.update(await self._apply_event_injects(event, *args, **kwargs))
        return True

    async def _apply_event_injects(self, event: EventType, *args, **kwargs) -> dict[str, Any]:
        add_kwargs = {}
        if event in self._injects:
            for callback in self._injects[event].sync_callback:
                add_kwargs.update(callback(*args, **kwargs))
            for callback in self._injects[event].async_callback:
                add_kwargs.update(await callback(*args, **kwargs))
        return add_kwargs

    async def _apply_global_injects(self, *args, **kwargs) -> dict[str, Any]:
        add_kwargs = {}
        for callback in self._global_injects.sync_callback:
            add_kwargs.update(callback(*args, **kwargs))
        for callback in self._global_injects.async_callback:
            add_kwargs.update(await callback(*args, **kwargs))
        return add_kwargs

    def global_event_inject(self, weight: int = 1) -> Callable[[InjectCallback], InjectCallback]:
        def decorator(func: InjectCallback):
            self.add_global_inject(func, weight)
            return func

        return decorator

    def add_global_inject(self, callback: InjectCallback, weight: int = 1) -> None:
        self._global_injects.add_callback(callback, weight)
        logger.debug(f"Global inject {callback.__name__} has been added, weight={weight}")

    def remove_global_inject(self, callback: InjectCallback) -> None:
        self._global_injects.remove_callback(callback)

    def event_inject(self, event: EventType, weight: int = 1) -> Callable[[InjectCallback], InjectCallback]:
        def decorator(func: InjectCallback):
            self.add_inject(event, func, weight)
            return func

        return decorator

    def add_inject(self, event: EventType, callback: InjectCallback, weight: int = 1) -> None:
        if event not in self._injects:
            self._injects[event] = EventCallbackContainer()
        self._injects[event].add_callback(callback, weight)
        logger.debug(f"Event inject {callback.__name__} has been added to event {event}, weight={weight}")

    def remove_inject(self, event: EventType, callback: InjectCallback) -> None:
        if event in self._injects:
            self._injects[event].remove_callback(callback)



from asyncio import iscoroutinefunction
from typing import Callable, Union

from loguru import logger

from .async_event_callback import AsyncEventCallback
from .event_callback import EventCallback
from .event_callback_factory import EventCallbackFactory
from .sync_event_callback import SyncEventCallback


class EventCallbackContainer:
    """
    A container class that stores callback functions
    """

    def __init__(self):
        self._sync_callback: list[SyncEventCallback] = []
        self._async_callback: list[AsyncEventCallback] = []

    def add_sync_callback(self, callback: SyncEventCallback) -> None:
        if callback not in self._sync_callback:
            logger.trace(f"Adding sync callback: {callback}")
            self._sync_callback.append(callback)
            self._sync_callback.sort(key=lambda item: item.weight, reverse=True)
        else:
            logger.trace(f"Callback already exists: {callback}")

    def add_async_callback(self, callback: AsyncEventCallback) -> None:
        if callback not in self._async_callback:
            logger.trace(f"Adding async callback: {callback}")
            self._async_callback.append(callback)
            self._async_callback.sort(key=lambda item: item.weight, reverse=True)
        else:
            logger.trace(f"Callback already exists: {callback}")

    def add_callback(self, callback: Union[EventCallback, Callable], weight: int = 1) -> None:
        if not isinstance(callback, EventCallback):
            callback = EventCallbackFactory.create(callback, weight)
        if isinstance(callback, AsyncEventCallback):
            self.add_async_callback(callback)
        elif isinstance(callback, SyncEventCallback):
            self.add_sync_callback(callback)
        else:
            raise TypeError(f'Callback type {type(callback)} not supported')

    def remove_sync_callback(self, callback: Union[SyncEventCallback, Callable]) -> None:
        if callback in self._sync_callback:
            logger.trace(f"Removing sync callback: {callback}")
            self._sync_callback.remove(callback)

    def remove_async_callback(self, callback: Union[AsyncEventCallback, Callable]) -> None:
        if callback in self._async_callback:
            logger.trace(f"Removing async callback: {callback}")
            self._async_callback.remove(callback)

    def remove_callback(self, callback: Union[EventCallback, Callable]) -> None:
        if not isinstance(callback, EventCallback):
            if iscoroutinefunction(callback):
                self.remove_async_callback(callback)
            else:
                self.remove_sync_callback(callback)
            return
        if isinstance(callback, AsyncEventCallback):
            self.remove_async_callback(callback)
        elif isinstance(callback, SyncEventCallback):
            self.remove_sync_callback(callback)
        else:
            raise TypeError(f'Callback type {type(callback)} not supported')

    def clear(self) -> None:
        self._sync_callback.clear()
        self._async_callback.clear()

    @property
    def sync_callback(self) -> list[SyncEventCallback]:
        return self._sync_callback

    @property
    def async_callback(self) -> list[AsyncEventCallback]:
        return self._async_callback



class MultipleError(Exception):
    def __init__(self, exceptions: list[Exception]):
        self._exceptions = exceptions
        self._info = f"There are {len(self._exceptions)} exceptions occurred. Use exceptions property for more details."

    @property
    def exceptions(self) -> list[Exception]:
        return self._exceptions

    def __str__(self) -> str:
        return self._info

    def __repr__(self) -> str:
        return self._info



from typing import Any, Callable, Generic, TypeVar

T = TypeVar('T', bound=Callable)


class EventCallback(Generic[T]):
    """
    A basic class that encapsulates a callback function
    """

    def __init__(self, callback: T, weight: int = 1) -> None:
        self._callback = callback
        self._weight = weight
        self._async = False

    @property
    def weight(self) -> int:
        return self._weight

    @weight.setter
    def weight(self, weight: int) -> None:
        self._weight = weight

    @property
    def callback(self) -> T:
        return self._callback

    @property
    def is_async(self) -> bool:
        return self._async

    def __eq__(self, __value: Any) -> bool:
        if isinstance(__value, self.__class__):
            return self._callback == __value._callback and self._weight == __value._weight
        else:
            return self._callback == __value

    def __str__(self) -> str:
        return f"EventCallback(callback={self._callback}, weight={self.weight}, async={self.is_async})"

    def __repr__(self) -> str:
        return str(self)



from typing import Awaitable, Callable, Type, Union

from loguru import logger

from .base_module import BaseModule
from ..event import EventType, EventCallbackContainer

FilterCallback: Type = Callable[..., Union[bool, Awaitable[bool]]]


class BusFilter(BaseModule):
    """
    The event bus filter module is responsible for filtering the event
    and determining whether to continue to propagate the event
    """

    def __init__(self):
        self._filters: dict[str, EventCallbackContainer] = {}
        self._global_filters: EventCallbackContainer = EventCallbackContainer()

    def clear(self):
        self._filters.clear()
        self._global_filters.clear()

    async def resolve(self, event: EventType, args, kwargs) -> bool:
        if await self._apply_global_filter(event, *args, **kwargs):
            return True
        if await self._apply_filter(event, *args, **kwargs):
            return True
        return False

    async def _apply_filter(self, event: EventType, *args, **kwargs) -> bool:
        if event in self._filters:
            for callback in self._filters[event].sync_callback:
                if callback(*args, **kwargs):
                    return True
            for callback in self._filters[event].async_callback:
                if await callback(*args, **kwargs):
                    return True
        return False

    async def _apply_global_filter(self, event: EventType, *args, **kwargs) -> bool:
        for callback in self._global_filters.sync_callback:
            if callback(event, *args, **kwargs):
                return True
        for callback in self._global_filters.async_callback:
            if await callback(event, *args, **kwargs):
                return True
        return False

    def global_event_filter(self, weight: int = 1) -> Callable[[FilterCallback], FilterCallback]:
        """
        Register to global filters by decorator\n
        Use decorator to register a global event filter function to the event bus,
        which can be asynchronous or synchronous.\n
        When passing a function, you can pass the weight at the same time,
        and the filter with the weight will be called first\n
        The filter return value:
            If it returns `True`, it means that this event is truncated and no longer propagates down.\n
            If it returns `False`, this event continues to propagate\n
        Example:
            @event_bus.on_global_event_filter()
            def message_logger(message, *_, **__):
                print(message)
            # Asynchronous functions also accepted
            @event_bus.on_global_event_filter()
            async def message_recoder(message, *_, **__):
                await ...
            # Weights can be passed simultaneously
            @event_bus.on_global_event_filter(10)
            async def message_filter(message, *_, **__):
                await ...

        :param weight: The selection weight of the filter
        :return: The decorator function
        """

        def decorator(func: FilterCallback):
            self.add_global_filter(func, weight)
            return func

        return decorator

    def add_global_filter(self, callback: FilterCallback, weight: int = 1) -> None:
        """
        Register for global filters\n
        Functions used to register to global filters inside, or can be used separately\n
        Please check **BusFilter.global_event_filter** for details\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.add_global_filter(message_logger)
            # Asynchronous functions also accepted
            async def message_recoder(message, *_, **__):
                await ...
            event_bus.add_global_filter(message_recoder)
            # Weights can be passed simultaneously
            async def message_filter(message, *_, **__):
                await ...
            event_bus.add_global_filter(message_filter, 10)

        :param callback: Event filter function
        :param weight: The selection weight of the filter
        """
        self._global_filters.add_callback(callback, weight)
        logger.debug(f"Global filter {callback.__name__} has been added, weight={weight}")

    def remove_global_filter(self, callback: FilterCallback) -> None:
        """
        Remove the global filter\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.remove_global_filter(message_logger)

        :param callback: Event filter function
        """
        self._global_filters.remove_callback(callback)

    def event_filter(self, event: EventType, weight: int = 1) -> Callable[[FilterCallback], FilterCallback]:
        """
        Register to event filters by decorator\n
        Use decorator to register an event filter function to the event bus,
        which can be asynchronous or synchronous.\n
        The event type can be a string, representing a custom type,
        or you can write an enum class that inherits the EnumEvent class,
        or write a class that inherits the AbstractEvent class for more customization. \n
        Please check **BusFilter.global_event_filter** for filter returns\n
        Example:
            @event_bus.event_filter('message_create')
            def message_logger(message, *_, **__):
                print(message)
            # Asynchronous functions also accepted
            @event_bus.event_filter('message_create')
            async def message_recoder(message, *_, **__):
                await ...
            # Weights can be passed simultaneously
            @event_bus.event_filter('message_create', 10)
            async def message_filter(message, *_, **__):
                await ...

        :param event: Event to filter to
        :param weight: The selection weight of the filter
        """

        def decorator(func: FilterCallback):
            self.add_filter(event, func, weight)
            return func

        return decorator

    def add_filter(self, event: EventType, callback: FilterCallback, weight: int = 1) -> None:
        """
        Register for event filters\n
        Functions used to register to event filters inside, or can be used separately\n
        Please check **BusFilter.event_filter** for details\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.add_filter('message_create', message_logger)
            @event_bus.event_filter('message_create')
            async def message_recoder(message, *_, **__):
                await ...
            event_bus.add_filter('message_create', message_recoder)
            # Weights can be passed simultaneously
            async def message_filter(message, *_, **__):
                await ...
            event_bus.add_filter('message_create', message_filter, 10)

        :param event: Event to filter to
        :param callback: Event filter function
        :param weight: The selection weight of the filter
        """
        if event not in self._filters:
            self._filters[event] = EventCallbackContainer()
        self._filters[event].add_callback(callback, weight)
        logger.debug(f"Event filter {callback.__name__} has been added to event {event}, weight={weight}")

    def remove_filter(self, event: EventType, callback: FilterCallback) -> None:
        """
        Remove the event filter\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.remove_filter('message_create', message_logger)

        :param event: Event to filter to
        :param callback: Event filter function
        """
        if event in self._filters:
            self._filters[event].remove_callback(callback)



from .base_bus import BaseBus
from .base_module import BaseModule
from .bus_filter import BusFilter
from .bus_inject import BusInject
from .module_exceptions import *

__ALL__ = [
    BaseBus,
    BaseModule,
    BusFilter,
    BusInject,
    MultipleError
]



from abc import ABC, abstractmethod
from asyncio import Semaphore, gather, get_event_loop, new_event_loop, run_coroutine_threadsafe, set_event_loop
from typing import Any, Awaitable, Callable, Type, Union

from loguru import logger

from .module_exceptions import MultipleError
from ..event import EventType, EventCallbackContainer

SubScriberCallback: Type = Callable[..., Union[Any, Awaitable[Any]]]


class BaseBus(ABC):
    """
    Event base class, which provides the most basic event subscription and triggering services.
    :param max_concurrent_tasks: The maximum number of tasks for an asynchronous task
    """

    def __init__(self, max_concurrent_tasks: int = 10):
        self._subscribers: dict[str, EventCallbackContainer] = {}
        self._semaphore = Semaphore(max_concurrent_tasks)
        self._raise_exception = False

    def on(self, event: EventType, *, weight: int = 1) -> Callable[[SubScriberCallback], SubScriberCallback]:
        """
        Subscribe to the event bus by decorator\n
        Use decorator to register an event handler to the event bus, which can be asynchronous or synchronous.\n
        The event type can be a string, representing a custom type,
        or you can write an enum class that inherits the EnumEvent class,
        or write a class that inherits the AbstractEvent class for more customization. \n
        Example:
            @event_bus.on('message_create')
            def message_logger(message, *_, **__):
                print(message)
            # Asynchronous functions also accepted
            @event_bus.on('message_create')
            async def message_recoder(message, *_, **__):
                await ...

        :param event: Event to subscribe to
        :param weight: The selection weight of the event handler
        :return: The decorator function
        """

        def decorator(func: SubScriberCallback):
            self.subscribe(event, func, weight=weight)
            logger.debug(f"{func.__name__} has subscribed to {event}, weight={weight}")
            return func

        return decorator

    def subscribe(self, event: EventType, callback: SubScriberCallback, *, weight: int = 1) -> None:
        """
        Subscribe to the event bus\n
        Functions used to subscribe to functions inside, or can be used separately\n
        Please check **BaseBus.on** for details\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.subscribe('message_create', message_logger)
            # Asynchronous functions also accepted
            async def message_recoder(message, *_, **__):
                await ...
            event_bus.subscribe('message_create', message_recoder)

        :param event: Event to subscribe to
        :param callback: Event callback function
        :param weight: The selection weight of the event handler
        """
        if event not in self._subscribers:
            self._subscribers[event] = EventCallbackContainer()
        self._subscribers[event].add_callback(callback, weight)

    def unsubscribe(self, event: EventType, callback: SubScriberCallback) -> None:
        """
        Unsubscribe from the event\n
        Example:
            def message_logger(message, *_, **__):
                print(message)
            event_bus.unsubscribe('message_create', message_logger)

        :param event: Event to subscribe to
        :param callback: Event callback function
        """
        if event in self._subscribers:
            self._subscribers[event].remove_callback(callback)

    def emit_sync(self, event: EventType, *args, **kwargs) -> None:
        """
        Trigger events in a blocking manner\n
        Example:
            emit_sync('message_create', "This is a message", user="Half")

        :param event: Event to be triggered
        """
        try:
            loop = get_event_loop()
            if loop and loop.is_running():
                future = run_coroutine_threadsafe(
                    self.emit(event, *args, **kwargs),
                    loop
                )
                future.result()
        except RuntimeError:
            loop = new_event_loop()
            set_event_loop(loop)
            try:
                loop.run_until_complete(
                    self.emit(event, *args, **kwargs)
                )
            finally:
                loop.close()
                set_event_loop(None)

    async def _run_with_semaphore(self, coroutine: Callable, *args, **kwargs):
        """
        Asynchronous function executor with limiter
        :param coroutine: Original asynchronous function
        """
        async with self._semaphore:
            return await coroutine(*args, **kwargs)

    @abstractmethod
    async def before_emit(self, event: EventType, *args, **kwargs) -> tuple[bool, dict]:
        return False, {}

    async def emit(self, event: EventType, *args, **kwargs) -> None:
        """
        Asynchronous trigger event\n
        Execution order:
            The greater the weight of the function, the more it is executed first,
            and the synchronous function takes precedence over the asynchronous function execution.\n
            Among them, the last asynchronous event handling function weight has no effect,
            because **asyncio.gather** will be used for concurrent execution,
            and the order of execution will be meaningless\n
        Example:
            await emit('message_create', "This is a message", user="Half")

        :param event: Event to be triggered
        """
        skip, extra_kwargs = await self.before_emit(event, *args, **kwargs)
        if skip:
            return
        kwargs.update(extra_kwargs)
        if event in self._subscribers:
            exceptions = []

            for callback in self._subscribers[event].sync_callback:
                try:
                    callback(*args, **kwargs)
                except Exception as e:
                    if self._raise_exception:
                        raise e
                    exceptions.append(e)

            async_handlers = [self._run_with_semaphore(callback, *args, **kwargs) for callback in
                              self._subscribers[event].async_callback]

            if self._raise_exception:
                await gather(*async_handlers, return_exceptions=False)
            else:
                results = await gather(*async_handlers, return_exceptions=True)
                if not (len(results) == 1 and results[0] is None):
                    exceptions.extend(results)

            if (exception_size := len(exceptions)) != 0:
                if exception_size == 1:
                    raise exceptions[0]
                else:
                    raise MultipleError(exceptions)

    def clear(self):
        self._subscribers.clear()

    @property
    def raise_exception_immediately(self) -> bool:
        return self._raise_exception

    @raise_exception_immediately.setter
    def raise_exception_immediately(self, value: bool) -> None:
        self._raise_exception = value



from typing import Any

from .event_callback import EventCallback, T


class AsyncEventCallback(EventCallback):
    """
    A class that encapsulates an async callback function
    """

    def __init__(self, callback: T, weight: int = 1):
        super().__init__(callback, weight)
        self._async = True

    async def __call__(self, *args, **kwargs) -> Any:
        return await self._callback(*args, **kwargs)
